var emptyFunction = function(){};
var DEFAULT_VALUES = {
	isVisible: true,
	readonly: false,
	aclIsVisible: true,
	aclReadOnly: false,
	render: emptyFunction,
	factoredWidgets: [],
	events: []
};

/**
* Creates an instance of Container
*
* @class
* @name Container
* @param {ScreenService} ScreenService - An instance of ScreenService.
*/
var Container = function(ScreenService) {
	var container = this;

	var _setDefaultValuesField = (function(container) {
		Util.extend(container, DEFAULT_VALUES);
		Util.extend(container, new Actionable());
		Util.extend(container, new Modeable());
		Util.extend(container, new Templatable());
	})(this);

	var _eachWidget = function(widget, callback) {
		var result = true;
		var callbackResult = callback(widget);
		if (callbackResult === false) {
			result = false;
		}
		if (widget.widgets && result !== false) {
			widget.widgets.every(function(widget) {
				var childResult = _eachWidget(widget, callback);
				if(childResult === false){
					result = false;
				}
				return childResult;
			});
		}
		return result;
	};

	/**
	* Executes a callback for every widget on the container and for its children
	* @function eachWidget
	* @memberof Container
	* @instance
	* @param {Function} callback - A function to be executed on the widget
	*/
	this.eachWidget = function(callback) {
		this.widgets.every(function(widget) {
			return _eachWidget(widget, callback);
		});
	};

	/**
	* Executes a callback for each factored widget on the container and for its children
	* @function eachFactoredWidget
	* @memberof Container
	* @instance
	* @param {Function} callback - A function to be executed on the widget
	*/
	this.eachFactoredWidget = function(callback) {
		this.factoredWidgets.forEach(function(widget) {
			_eachWidget(widget, callback);
		});
	};

	var _getActiveWidget = function _getActiveWidget() {
		var widget;

		container.eachWidget(function(candidateWidget) {
			if (candidateWidget.active) {
				widget = candidateWidget;
				return false;
			}
		});

		if (!widget) {
			container.eachFactoredWidget(function(candidateWidget) {
				if (candidateWidget.active) {
					widget = candidateWidget;
					return false;
				}
			});
		}

		return widget;
	};

	/**
	* Retrives the active widget on the container
	* @function getActiveWidget
	* @memberof Container
	* @instance
	* @returns {Widget} Active widget
	*/
	this.getActiveWidget = function getActiveWidget() {
		var widget = null;
		widget = _getActiveWidget();

		if (!widget) {
			widget = this.widgets[0];
		}

		return widget;
	};

	/**
	* Updates the checking state on the widget rows
	* @function updateCheckedRows
	* @memberof Container
	* @instance
	*/
	this.updateCheckedRows = function() {
		var widget = ScreenService.isPopupOpen() ? ScreenService.getPopupOpen() : this.getActiveWidget();
		if (widget && widget.dataSource && widget.getCheckedRows && widget.getCheckedRows().length &&
			widget.dataSource.data && widget.dataSource.data.length) {
			this.widgetToSelect = widget;

			var qtCheckedRows = widget.getCheckedRows().length;
			var rowText = qtCheckedRows > 1 ? " rows" : " row";

			this.textSelectedItens = qtCheckedRows + ScreenService.i18n(rowText + " selected");
		}

		if (widget && widget.updateCheckingState) {
			widget.updateCheckingState(this.textSelectedItens);
		}
	};

	var _inactiveWidget = function(widget, exceptionWidget) {
		if(widget !== exceptionWidget && widget.deactivate){
			widget.deactivate();
		}
		if (widget.widgets) {
			widget.widgets.forEach(function(childWidget) {
				_inactiveWidget(childWidget, exceptionWidget);
			});
		}
		if(widget.fields){
			widget.fields.forEach(function(childField){
				if(childField.selectWidget){
					_inactiveWidget(childField.selectWidget, exceptionWidget);
				}	
			});
		}
	};

	/**
	* Deactivates all the widgets, leaving only the exception active
	* @function inactiveWidgets
	* @memberof Container
	* @instance
	* @param {Widget} exceptionWidget - Widget to be left active
	*/
	this.inactiveWidgets = function(exceptionWidget) {
		this.widgets.forEach(function(widget) {
			_inactiveWidget(widget, exceptionWidget);
		});
	};

	/**
	* Asynchronously refreshes the container
	* @function refresh
	* @memberof Container
	* @instance
	*/
	this.refresh = function() {
		setTimeout(function() {
			if (container.scrollers) {
				container.scrollers.forEach(function(scroller) {
					scroller.refresh();
				});
			}
		}, 1000);
	};
	this.getParent = function() {
		return null;
	};
	var getWidgetFlow = {
		'object': function(widget) {
			return widget;
		},
		'number': function(widget) {
			return container.widgets[widget];
		},
		'string': function(widget) {
			var widgetResult;
			try {
				container.eachWidget(function(itemWidget) {
					if (itemWidget.name == widget) {
						widgetResult = itemWidget;
						throw 'Stop search';
					}
				});
			} catch (e) {}
			return widgetResult;
		},
		'undefined': function() {
			return container.widgets[0];
		}
	};

	/**
	* Retrieves the widget specified
	* @function getWidget
	* @memberof Container
	* @instance
	* @param {Widget|Number|String} [widget] - Specific characteristic of the widget. If undefined, returns all the widgets
	* @returns {Widget} Widget(s) with the specified characteristic(s)
	*/
	this.getWidget = function(widget) {
		return getWidgetFlow[typeof widget](widget);
	};

	/**
	* Verify if the active widget has checked rows
	* @function hasCheckedRows
	* @memberof Container
	* @instance
	*/
	this.hasCheckedRows = function() {
		return this.getActiveWidget()._hasCheckedRows();
	};

	/**
	* Verify if the active widget has checked rows without using cached data
	* @function hasCheckedRowsWithoutCache
	* @memberof Container
	* @instance
	*/
	this.hasCheckedRowsWithoutCache = function() {
		return this.getActiveWidget()._hasCheckedRowsWithoutCache();
	};

	/**
	* Retrieves the actions of the container
	* @function getActions
	* @memberof Container
	* @instance
	*/
	this.getActions = this.getActionsToFilter;

	/**
	* Sets the property 'actions' of the container with the proper actions from the widgets
	* @function prepareActions
	* @memberof Container
	* @instance
	*/
	this.prepareActions = function() {
		// To concat actions only one time
		if (!this.actions) {
			this.actions = [];
			this.widgets.forEach(function(widget) {
				container.actions = container.actions.concat(widget.getActions());
			});
		}
	};

	/**
	* Verifies if the action is visible
	* @function addictionalFilterAction
	* @memberof Container
	* @instance
	* @param {Action} Action to be verified
	* @returns {Boolean} True if action is visible
	*/
	this.addictionalFilterAction = function(action) {
		return (action.widget.active || this.showAllWidgetActions);
	};

	/**
	* Selects all the rows of the widget
	* @function selectAll
	* @memberof Container
	* @instance
	* @param {Widget|Number|String} [widget] - Specific characteristic of the widget that will have its rows selected
	*/
	this.selectAll = function(widget) {
		this.getWidget(widget).selectAll();
	};

	/**
	* Unselects all the rows of the widget
	* @function unSelectAll
	* @memberof Container
	* @instance
	* @param {Widget|Number|String} [widget] - Specific characteristic of the widget that will have its rows unselected
	*/
	this.unSelectAll = function(widget) {
		this.getWidget(widget).unSelectAll();
	};

	/**
	* Selects / Unselects the rows on the widget
	* @function toogleSelection
	* @memberof Container
	* @instance
	* @param {Widget|Number|String} [widget] - Specific characteristic of the widget that will be toogled
	*/
	this.toogleSelection = function(widget) {
		if (this.getWidget(widget).hasCheckedRows()) {
			this.unSelectAll(widget);
		} else {
			this.selectAll(widget);
		}
	};

	function reloadLazyDataSource (item, initialFilter) {
		if (item.dataSource && !item.dataSource.lazyLoad && (!item.dataSource.data || !item.dataSource.data.length || (initialFilter && initialFilter.length))) {
			item.reload(initialFilter).then(function() {
				if (item.initDataSource) {
					item.initDataSource();
				}
			});
		}
	}

	/**
	* Reloads the lazy data sources of the widgets on the container
	* @function reloadLazyDataSources
	* @memberof Container
	* @instance
	* @param {Object[]} [maintainWindowFilter] - List of filters to be applied when reloading
	*/
	this.reloadLazyDataSources = function (maintainWindowFilter) {
		container.widgets.forEach(function(widget) {
			reloadLazyDataSource(widget, Util.isUndefinedOrNull(maintainWindowFilter) ? null : maintainWindowFilter[widget.name]);
			widget.fields.forEach(function(field) {
				reloadLazyDataSource(field);
			});
		});
	};

};